% Program accompanying Book Image and Video Compression: Fundamentals,
% techniques, and applications by Madhuri A Joshi, Mehul S Raval, Yogesh
% Dandawate, Kalyani R Joshi, Shilpa Metkar, CRC Press, Nov. 2014
% Contact email: mehul.raval@gmail.com
% This the program with Chapter 3 Image Transforms
% The program embeds and extracts a binary watermark into an DC and AC
% coefficient of an 512 x 512 grayscale image using quantization index modulation

close all; %close all figures;
clear all; % Clear the workspace

I=imread('3.12a.jpg');% Read the Cover image
I=imresize(I,[512 512]);
subplot(121);
imshow(I);
title('Original Image');
[rows cols] = size(I); % Get the image size
IW=zeros(rows,cols);%Initialize Watermarked Image to zero

rows=rows/8; %Block index 
cols=cols/8;%Block index 
W = randi([0,1],1,2*rows*cols); % Initialize Watermark

dmin=1 % QIM step size
k = 1; % initialize the watermark column index

% Watermark embedding
for i=1:rows
    for j=1:cols
		block = I((i-1)*8+1:i*8,(j-1)*8+1:j*8); % get a 8x8 block
        dctblock = dct2(block); % take its dct
        p=dctblock(1,1);% Select DC coefficient
        % Embed Watermark in DC coefficient
        if W(1,k)== 1
            q = p+dmin-mod(p,2*dmin);
        else
            q = p+dmin-mod(p+dmin,2*dmin);
        end
        dctblock(1,1)=q;
        k=k+1; %Point to next watermark bit
        
        p=dctblock(2,1);% Select AC coefficient
        % Embed Watermark in AC coefficient
        if W(1,k)== 1
            r = p+dmin-mod(p,2*dmin);
        else
            r = p+dmin-mod(p+dmin,2*dmin);
        end
        dctblock(2,1)=r;
        k=k+1;
        hblock = idct2(dctblock);% Apply Inverse DCT 
        IW((i-1)*8+1:i*8,(j-1)*8+1:j*8) = hblock;% Replace the block
    end
end
subplot(122);
imshow(uint8(IW));% Display Watermarked Image
title('Watermarked Image');

% Watermark Decoding
[rows cols] = size(IW)% Determine the size of Watermarked Image
rows=rows/8;%Block index
cols=cols/8;%Block index
wr = randi([0,0],1,2*rows*cols); % Initialize recovered block to zero
k=1; %Watermark column index
for i=1:rows
    for j=1:cols
		blockw = IW((i-1)*8+1:i*8,(j-1)*8+1:j*8); % get a 8x8 block
        dctblockw = dct2(blockw); % take its dct
        pw=round(dctblockw(1,1));% Select DC coefficient
        if mod(pw,2*dmin)==0 % If coefficient is even
        wr(1,k)=0; % Watermark bit is 0
        else
        wr(1,k)=1; %Watermark bit is 1
        end
        k=k+1; %Point to next watermark index
        
        pw=round(dctblockw(2,1));% Select AC coefficient
        if mod(pw,2*dmin)==0 % If coefficient is even
        wr(1,k)=0; % watermark bit is 0
        else
        wr(1,k)=1; %watermark bit is 1
        end
        k=k+1;
    end
end

%Compare embedded and extracted watermark
isequal(W,wr)
            
