% Program accompanying Book Image and Video Compression: Fundamentals,
% techniques, and applications by Madhuri A Joshi, Mehul S Raval, Yogesh
% Dandawate, Kalyani R Joshi, Shilpa Metkar, CRC Press, Nov. 2014
% Contact email: mehul.raval@gmail.com
% This the program with Chapter 8 Compressive Sensing
% The program embeds and extracts a binary watermark using compressive
% sensing framework. This program needs solver available at
% http://users.ece.gatech.edu/~justin/l1magic
% Written by: Justin Romberg, Caltech
% Email: jrom@acm.caltech.edu
% Created: October 2005


close all; %close all figures;
clear all; % Clear the workspace

% Generate the watermark
% Watermark length
N = 100;
% number of nonzero values in the watermark
NZ = 20;
% number of observations to make
K = 20480;

% Generate random watermark
x = zeros(N,1);
x1 = randperm(N);
x(x1(1:NZ)) = randn(NZ,1);

%Generate random measurement matrix psi
psi = randn(K,N)/100;

%Generate random measurement
y=psi*x;

% Insert the watermark in image
I=imread('Fig.8.21a.jpg');% Read the Cover image
I=imresize(I,[256 256]);
subplot(121);
imshow(I);
title('Original Image');
[rows cols] = size(I); % Get the image size
% IW=zeros(rows,cols);%Initialize Watermarked Image to zero

r=rows/8; %Block index 
c=cols/8;%Block index 

%Basic JPEG Quantization matrix with quality factor of 50
QM=[17 12 11 17 25 41 52 62;
       13 13 15 20 27 59 61 56;
       15 14 17 25 42 58 70 57;
       15 18 23 30 52 88 81 63;
       19 23 38 57 69 110 104 78;
       25 36 56 65 82 105 114 93;
       50 65 79 88 104 122 121 102;
       73 93 96 99 113 101 104 100];
   
   k=1;%Index to measurement vector
% Watermark embedding
for i=1:r
    for j=1:c
        block = I((i-1)*8+1:i*8,(j-1)*8+1:j*8); % get a 8x8 image block
        dctblock = dct2(block); % take its dct
        dctblock=dctblock./QM;% Divide the dctblock by quantization matrix
        % Add 20 middle frequency dct coefficients and y
        dctblock(1,2)=dctblock(1,2)+y(k,1);k=k+1;
        dctblock(1,3)=dctblock(1,3)+y(k,1);k=k+1;
        dctblock(1,4)=dctblock(1,4)+y(k,1);k=k+1;
        dctblock(1,5)=dctblock(1,5)+y(k,1);k=k+1;
        dctblock(1,6)=dctblock(1,6)+y(k,1);k=k+1;
        dctblock(2,1)=dctblock(2,1)+y(k,1);k=k+1;
        dctblock(2,2)=dctblock(2,2)+y(k,1);k=k+1;
        dctblock(2,3)=dctblock(2,3)+y(k,1);k=k+1;
        dctblock(2,4)=dctblock(2,4)+y(k,1);k=k+1;
        dctblock(2,5)=dctblock(2,5)+y(k,1);k=k+1;
        dctblock(3,1)=dctblock(3,1)+y(k,1);k=k+1;
        dctblock(3,2)=dctblock(3,2)+y(k,1);k=k+1;
        dctblock(3,3)=dctblock(3,3)+y(k,1);k=k+1;
        dctblock(3,4)=dctblock(3,4)+y(k,1);k=k+1;
        dctblock(4,1)=dctblock(4,1)+y(k,1);k=k+1;
        dctblock(4,2)=dctblock(4,2)+y(k,1);k=k+1;
        dctblock(4,3)=dctblock(4,3)+y(k,1);k=k+1;
        dctblock(5,1)=dctblock(5,1)+y(k,1);k=k+1;
        dctblock(5,2)=dctblock(5,2)+y(k,1);k=k+1;
        dctblock(6,1)=dctblock(6,1)+y(k,1);k=k+1;
        dctblock=dctblock.*QM; % Multiplication by Quantization matrix
        IW((i-1)*8+1:i*8,(j-1)*8+1:j*8)=idct2(dctblock);% Perform inverse DCT
    end
end

%Display watermark image
subplot(122);
imshow(uint8(IW));% Display Watermarked Image
title('Watermarked Image');
k=1;%Recovered measurement index
% Watermark extraction
for i=1:r
    for j=1:c
        block = IW((i-1)*8+1:i*8,(j-1)*8+1:j*8); % get a 8x8 image block
        dctblock = dct2(block); % take its dct
        dctblock=dctblock./QM;% Divide the dctblock by quantization matrix
        yr(k,1)=dctblock(1,2);k=k+1;
        yr(k,1)=dctblock(1,3);k=k+1;
        yr(k,1)=dctblock(1,4);k=k+1;
        yr(k,1)=dctblock(1,5);k=k+1;
        yr(k,1)=dctblock(1,6);k=k+1;
        yr(k,1)=dctblock(2,1);k=k+1;
        yr(k,1)=dctblock(2,2);k=k+1;
        yr(k,1)=dctblock(2,3);k=k+1;
        yr(k,1)=dctblock(2,4);k=k+1;
        yr(k,1)=dctblock(2,5);k=k+1;
        yr(k,1)=dctblock(3,1);k=k+1;
        yr(k,1)=dctblock(3,2);k=k+1;
        yr(k,1)=dctblock(3,3);k=k+1;
        yr(k,1)=dctblock(3,4);k=k+1;
        yr(k,1)=dctblock(4,1);k=k+1;
        yr(k,1)=dctblock(4,2);k=k+1;
        yr(k,1)=dctblock(4,3);k=k+1;
        yr(k,1)=dctblock(5,1);k=k+1;
        yr(k,1)=dctblock(5,2);k=k+1;
        yr(k,1)=dctblock(6,1);k=k+1;
    end
end

% Watermark recovery
x0 = inv(psi'*psi)*psi'*yr;%Initial estimate
xr = l1decode_pd(x0, psi, [], yr, 1e-4, 30);%recovered watermark
recoveryerr=mse(x-xr);% Recovery error 

